-- returns a vector where all components are multiplied by the specified amount
---@param amount number
---@return Vector
Vector.amplify = function(self, amount)
    return Vector(self.x * amount, self.y * amount, self.z * amount)
end

---@class modes
local MODES = {
    RANDOM = 'random',
    MANUAL = 'manual'
}
-- amount of tiles camera should move per second (at zoom level 1)
local BASE_SPEED = 2
-- minimum distance between new and current destination
local MIN_DISTANCE = 10

-- sidebar button icon
local previewFrame
-- city size
local cityHeight
-- current zoomlevel of city
local scale

---@class Camera contains information used for camera movement
---@field location vector current camera location
---@field targetDistance number the distance between the camera's current position and the target position
---@field frameSpeed vector the distance moved per frame
---@field target vector the target position for the camera
---@field speed number the speed multiplier set by the user
local camera = {}

---@class Settings all plugin settings
---@field enabled boolean whether the camera tool is currently running
---@field firstStart boolean whether its the first time starting the tool
---@field mode modes the mode in which camera destinations should be selected
---@field speedMultiplier string
---@field destinations Vector[] destinations for the camera in manual mode
local settings

function script:init()
    previewFrame = script:getDraft():getPreviewFrame()
    --get settings table
    settings = TheoTown.getStorage()['$cinematicam00:settings'] or {}
    settings.speedMultiplier = settings.speedMultiplier or '1'
    settings.mode = settings.mode or MODES.RANDOM
    settings.firstStart = settings.firstStart
    if settings.firstStart == nil then
        settings.firstStart = true
    end
    settings.enabled = false
    -- check if speed multiplier is a valid number
    if not pcall(function ()
        camera.speed = BASE_SPEED * tonumber(settings.speedMultiplier)
    end) then
        -- reset it to the default
        settings.speedMultiplier = '1'
        camera.speed = BASE_SPEED
    end
    math.randomseed(os.time())
end

function script:enterCity()
    -- if they ever add rectangular cities i have a problem
    cityHeight = City.getHeight()
    settings.enabled = false
    settings.destinations = City.getStorage()['$cinematicam00:destinations'] or {}
end

function script:exitCity()
    settings.enabled = false
end

-- calculates distance to target and sets stepX and stepY accordingly
---@param target vector the desired target location
local function setDestination(target)
    -- calculate distance to target
    local targetVector = target - camera.location
    camera.targetDistance = targetVector:length()
    -- calculate time of travel in seconds
    local timeOfTravel = camera.targetDistance / camera.speed
    -- calculate distance that needs to be traveled per second
    camera.frameSpeed = targetVector / timeOfTravel
end

-- only applicable in manual point mode
local destinationIndex
local function nextDestination()
    -- get the next destination in manual mode
    local function manual()
        if destinationIndex > #settings.destinations then
            destinationIndex = 1
        end
        local destination = settings.destinations[destinationIndex]
        destinationIndex = destinationIndex + 1
        setDestination(destination)
    end

    -- get a random destination at least MIN_DISTANCE tiles away from the current location
    local function random()
        ---@param current number the current location of the camera on an axis
        local function getRandomPoint(current)
            local randomPos = math.random(cityHeight - MIN_DISTANCE * 2)
            if randomPos >= current - MIN_DISTANCE then
                randomPos = randomPos + MIN_DISTANCE * 2
            end
            return randomPos
        end

        camera.target = Vector(getRandomPoint(camera.location.x), getRandomPoint(camera.location.y))
        setDestination(camera.target)
    end

    if settings.mode == MODES.RANDOM then
        random()
    elseif settings.mode == MODES.MANUAL then
        manual();
    end
end

-- creates the dialog with tutorials shown on first startup
local function showFirstStartDialog()
    local dialog = GUI.createDialog{
        icon = Icon.SETTINGS,
        title = 'cinematic camera tool tutorial',
        width = 400,
        height = 180,
        actions = {
            {
                icon = Icon.OK,
                text = 'lets start',
                autoClose = true,
                golden = true
              },
        }
    }
    local layout = dialog.content:addLayout{vertical = true}
    --totally not stolen from theotown forums
    local function addLine(height)
        local line = layout:addLayout{height = height}
        return line
    end
    local read1, read2, read3
    addLine(30):addButton{
        width = 28,
        icon = Icon.MENU,
        text = 'settings',
        isPressed = function () return read1 end,
        onClick = function ()
            read1 = true
            GUI.createDialog{
                icon = Icon.MENU,
                title = 'settings',
                text = 'the plugin adds a couple settings to your settings menu, there you can adjust the camera speed and select what mode you want',
            }
        end
    }
    addLine(30):addButton{
        width = 28,
        icon = previewFrame,
        text = 'random mode',
        isPressed = function () return read2 end,
        onClick = function ()
            read2 = true
            GUI.createDialog{
                icon = previewFrame,
                title = 'random mode',
                text = 'in random mode the camera will pan to random spots on your map',
            }
        end
    } 
    addLine(30):addButton{
        width = 28,
        icon = Draft.getDraft('$cinematicamTool00'):getPreviewFrame(),
        text = 'manual mode',
        isPressed = function () return read3 end,
        onClick = function ()
            read3 = true
            GUI.createDialog{
                icon = previewFrame,
                title = 'manual mode',
                text = 'in manual mode a button will be added to your toolbar, with this tool you can add waypoints to any tile in your city. When the camera is activated it will pan to each waypoint in the order you placed them',
            }
        end
    }

    local line1 = addLine(15)
    line1:addLabel{
        text = 'welcome to the tutorial for the cinematic camera tool'
    }
    local line2 = addLine(15)
    line2:addLabel{
        text = 'click the buttons to view info about the various features'
    }
end

local function openDestinationTool()
    GUI.createDialog{title='just one more thing', h=120,  text='you should create some destinations for the camera to travel to first', actions={icon = Icon.OK, text = Translation.control_ok}}
    City.createDraftDrawer(Draft.getDraft('$cinematicamTool00')).select()
end

local function setupFixedMode()
    if #settings.destinations < 2 then
        openDestinationTool()
        return false
    end
    destinationIndex = 1
    return true
end

---@type GUI
local button
local function createSideBarButton()
    local sidebar = GUI.get('sidebarLine')
    local size = sidebar:getFirstPart():getChild(2):getWidth()
    -- add camera button to the sidebar
    button = sidebar:getFirstPart():addButton{
        width = size,
        height = size,
        icon = previewFrame,
        frameDefault = Icon.NP_BLUE_BUTTON,
        frameDown = Icon.NP_BLUE_BUTTON_PRESSED,
        onClick = function ()
            if not settings.enabled then
                camera.location = Vector()
                camera.location.x, camera.location.y, scale = City.getView()
                if settings.mode == MODES.MANUAL then
                    if not setupFixedMode() then return end
                end
                nextDestination()
            end
            -- enable camera movement in script:update
            settings.enabled = not settings.enabled
        end,
        isPressed = function () return settings.enabled end
    }
end

function script:buildCityGUI()
    createSideBarButton()
end

---@type number
local lastMs
-- gets how often script:update gets called per second
---@return number
local function updatesPerSecond()
    local ms = Runtime.getTime()
    if lastMs == nil then
        lastMs = ms
        return -1
    else
        local fps = 1000 / (ms - lastMs)
        lastMs = ms
        return fps
    end
end

---@param fps number
local function stepCamera(fps)
    if camera.targetDistance < 0 then
        nextDestination()
    end
    -- super fancy magic formula for turning zoom level into a nice camera speed
    local multiplier = 2 / (scale + 1)
    multiplier = multiplier / fps

    -- reduce distance to target
    camera.targetDistance = camera.targetDistance - camera.speed * multiplier
    -- move camera
    camera.location = camera.location + camera.frameSpeed:amplify(multiplier)

    City.setView(camera.location.x, camera.location.y)
end

function script:update()
    if settings.enabled then
        -- get view to adjust for current zoom level
        _, _, scale = City.getView()

        local fps = updatesPerSecond()
        if fps == -1 then return end
        stepCamera(fps)
    end
    -- hide the button if ui is hidden
    button:setVisible(not TheoTown.SETTINGS.hideUI)
end

function script:settings()
    return {
        {
            name = 'speed multiplier',
            desc = 'Enter the speed multiplier for the camera here (default is 1)',
            value = settings.speedMultiplier,
            onChange = function(newState)
                pcall(function ()
                    camera.speed = BASE_SPEED * tonumber(newState)
                    settings.speedMultiplier = newState
                end)
                settings.enabled = false
            end,
            onReset = function()
                settings.speedMultiplier = '1'
                camera.speed = BASE_SPEED
                settings.enabled = false
            end
        },
        {
            name = 'mode',
            value = settings.mode,
            values = {MODES.MANUAL, MODES.RANDOM},
            onChange = function (newState)
                settings.mode = newState
                settings.enabled = false
            end,
            onReset = function()
                settings.mode = MODES.RANDOM
                settings.enabled = false
            end
        },
        {
            name = 'show first start dialog again',
            value = false,
            onChange = function ()
                settings.firstToolStart = true
                Runtime.postpone(function() showFirstStartDialog() end, 100)
            end
        }
    }
end