---@class modes
local MODES = {
    RANDOM = 'random',
    MANUAL = 'manual'
}
---@class Settings
---@field firstToolStart boolean
---@field toolEnabled boolean whether the user is currently using the tool
local settings
local markerFrame
---@type draft
local draft

-- gets all destinations from city storage
local function getDestinations()
    return City.getStorage()['$cinematicam00:destinations'] or {}
end

function script:init()
    settings = TheoTown.getStorage()['$cinematicam00:settings'] or {}
    settings.firstToolStart = settings.firstToolStart or true
    settings.toolEnabled = false
    draft = script:getDraft()
    markerFrame = draft:getFrame()
end

---@type GUI
local button
local function addSideBarButton()
    local sidebar = GUI.get('sidebarLine')
    local size = sidebar:getFirstPart():getChild(2):getWidth()
    button = sidebar:getFirstPart():addButton{
        width = size,
        height = size,
        icon = script:getDraft():getPreviewFrame(),
        frameDefault = Icon.NP_BLUE_BUTTON,
        frameDown = Icon.NP_BLUE_BUTTON_PRESSED,
        onClick = function ()
            if not settings.toolEnabled then
                City.createDraftDrawer(draft):select()
            end
        end,
        isPressed = function () return settings.toolEnabled end
    }
end

function script:enterCity()
    settings.destinations = getDestinations()
end

function script:exitCity()
    settings.toolEnabled = false
end

function script:buildCityGUI()
    -- add camera button to the sidebar
    if settings.mode == MODES.MANUAL then
        addSideBarButton()
    end
end

---@param destination vector
---@param index number
local function drawDestinationMarker(destination, index)
    Drawing.setTile(destination.x, destination.y, 5, 3)
    Drawing.drawImage(markerFrame)
    Drawing.setTile(destination.x, destination.y, 16, -22)
    Drawing.drawText(index, 0, 0, nil, 0.5, 0.5)
    Drawing.reset()
end

function script:drawCity()
    if settings.toolEnabled == true then
        for i = 1, #settings.destinations, 1 do
            local destination = settings.destinations[i]
            drawDestinationMarker(destination, i)
        end
    end
end

local function showFirstStartDialog()
    GUI.createDialog{
        title = 'waypoint tool',
        text = 'one more thing, click on any tile to place a waypoint, click again to remove it. Click reset to remove all waypoints',
        actions = {
            {
                icon = Icon.OK,
                text = 'lets start',
                autoClose = true,
                golden = true
              },
        }
    }
end

function script:event(x, y, level, event)
    if event == script.EVENT_TOOL_ENTER then
        if settings.firstToolStart then
            showFirstStartDialog()
            settings.firstToolStart = false
        end
        settings.toolEnabled = true
        TheoTown.registerToolAction{
            icon = Icon.REMOVE,
            name = 'reset',
            onClick = function()
                if settings.enabled then return end
                settings.destinations = {}
            end
        }
        settings.toolEnabled = true
    elseif event == script.EVENT_TOOL_LEAVE then
        settings.toolEnabled = false
    end
end

local function hasDestination(x, y)
    for i = 1, #settings.destinations, 1 do
        local endPoint = settings.destinations[i]
        if endPoint.x == x and endPoint.y == y then
            return true, i
        end
    end
    return false
end

function script:click(x, y)
    local exists, index = hasDestination(x, y)
    if exists then
        table.remove(settings.destinations, index)
    else
        table.insert(settings.destinations, Vector(x, y))
    end
end

function script:update()
    -- hide the button if ui is hidden
    if button ~= nil then
        button:setVisible(not TheoTown.SETTINGS.hideUI)
    end
end
